function imgLib( file, columns, rows, xsize, ysize )  -- Library object factory
  local self = { iC = columns, iX = xsize, iY = ysize }
  self.iO = image.load(file)
  self.iQ = rows * columns

  local draw = function (x, y, i, ...)  -- Draw picture and border if specified
    if self.iO and i >= 0 and i < self.iQ then
      local sX, sY, vA = self.iX * (i % self.iC), self.iY * (math.floor(i / self.iC)), ...
      self.iO:draw(x-sX, y-sY, x, y, self.iX, self.iY)
      if type (vA) == "number" then
        screen.drawrect(x, y, x + self.iX - 1, y + self.iY - 1, vA)
      end
    end
  end

  local close = function ()  --Close image
    self.iO:close()
  end

  return {draw = draw, close = close}
end

board = {} -- board object manages display

function board:cX ( index )
  return self.bX + self.tP * ((index - 1) % self.tX)
end

function board:cY ( index )
  return  self.bY + self.tP * math.floor((index - 1) / self.tX)
end

function board:shadow ( tile) --draw the shadow
  screen.fillrect(self:cX(tile)+self.sO, self:cY(tile)+self.sO, self.tS, self.tS, self.cS)
end

function board:up ( tile, index )  -- Show the card face image.
  local x,y = self:cX(tile), self:cY(tile)
  self.iLO.draw(x, y, index, self.cBo)
end

function board:down ( tile )  -- Show the card back.
  self.cBa:draw(self:cX(tile), self:cY(tile), self.tS, self.tS)
end

function board:off ( tile )  -- Remove the card.
  self.bG:draw(0, 0, self:cX(tile), self:cY(tile), self.tP, self.tP)
end

function board:pick ( x, y )  -- Return index number from touch position.
  if x >= self.bX and y >= self.bY and x < self.eX and y < self.eY then  -- Exclude border.
    local dx, dy = x - self.bX, y - self.bY  -- Relative to playing area.
    if dx % self.tP < self.tS and dy % self.tP < self.tS then  -- Exclude spaces between tiles.
      return 1 + self.tX * math.floor(dy / self.tP) + math.floor(dx / self.tP)
    else
      return nil
    end
  else
    return nil
  end
end

function board:new ( tileSize, minBorder, spaceSize, imageLibObject, background, cardBack, cardBorder, cardShadow)
  self.tS = tileSize
  self.iLO = imageLibObject
  self.bG = background
  self.cBa = cardBack
  self.cBo = cardBorder
  self.cS = cardShadow
  local sX, sY = screen.width(), screen.height() -- Just to initialize other variables
  self.sO = math.floor (spaceSize / 2)  -- Shadow offset
  self.tP = tileSize + spaceSize
  self.tX = math.floor((sX - 2 * minBorder + spaceSize) / self.tP)
  self.tY = math.floor((sY - 2 * minBorder + spaceSize) / self.tP)

  if (self.tY * self.tX) % 2 == 1 then  --Odd number of tiles must be adjusted
   if self.tY > self.tX then
     self.tY = self.tY - 1
   else
     self.tX = self.tX - 1
   end
  end
  self.bX = math.floor((sX - self.tX * self.tP + spaceSize) / 2) -- Left margin
  self.bY = math.floor((sY - self.tY * self.tP + spaceSize) / 2) -- Top margin
  self.eX = sX - self.bX - 1 -- Right limit
  self.eY = sY - self.bY - 1 -- Bottom limit
  return self.tX * self.tY
end

function board:set()
  self.bG:draw(0, 0)
--  screen.fillrect(0,0,screen.width(),screen.height(),self.bG) --Background
  for i = 1, self.tX * self.tY, 1 do  --Set the board
    self:shadow(i)
    self:down(i)
  end
  screen.update()
end

function noteHz (n) -- Convert note number to Hz
  return 2 ^ (n / 12) * 27.5
end

function soundText(state)
  text.size(16)
  text.color(BORDER)
  local btn = BUTTONS.sound
  if state then
    s = " "
  else
    s = " "
  end
  bkg:draw( 0, 0, btn.x, btn.y, btn.w, btn.h )
  text.draw(0, 160, s, "center", 400)
  screen.update()
end

function soundToggle ()
  soundON = not soundON
  soundText(soundON)
end

-- Constants and metaconstants
SHADOW = color.new(0, 85, 170) -- Tile edge shadow
HIGHLIGHT = color.new(255, 170, 85, 128)  -- Button highlight.
BORDER = color.new(222, 222, 222) -- Edge of tile when turned
TILE_SPACE = 4  --Space between tiles
BASE_NOTE = 45 -- Lowest beep to use
SHUF_PASS = 3  -- How many times to shuffle
BORDER_MIN = 10  -- Minimum blank border
WIN_WAV = "bababee.wav"
CARD_ROWS = 4 -- How many images in the collection?
CARD_COLS = 8
CARD_QTY = CARD_ROWS * CARD_COLS
BUTTONS = {
  easy = { x = 62, y = 109, w = 128, h = 52 },
  hard = { x = 205, y = 114, w = 134, h = 42 },
  sound = { x = 155, y = 160, w = 90, h = 18}
  } -- Screen coordinates of option buttons.
OPTIONS = {
  easy = { tile = 50, file = "tiles_l.png", back = "card_l.png", wait = 400 },
  hard = { tile = 40, file = "tiles_s.png", back = "card_s.png", wait = 300 }
  } -- Game options for each selection.
ACTIONS = {
  sound = "soundToggle()"
  }

menu = image.load("welcome.png")
bkg = image.load("back.png")
soundON = true

repeat -- Main loop
-- Draw the options menu.
  menu:draw(0, 0)
  soundText(soundON)
  screen.update()
  choice = nil
  die = false

  repeat  -- Select difficulty
    control.read(1)
    if control.isTouch() == 1 and (touch.click() == 1 or touch.up() == 1) then
      xT, yT = touch.pos()
      for b, d in pairs(BUTTONS) do
        if xT >= d.x and yT >= d.y and xT < d.x + d.w and yT < d.y + d.h then  -- Button selected
          screen.fillrect(d.x, d.y, d.w, d.h, HIGHLIGHT) -- Highlight selected button.
          screen.update()
          os.wait(100)
          choice = OPTIONS[b]  -- Get parameters for selected game.
          if ACTIONS[b] then
            assert(loadstring(ACTIONS[b]))()
          end
        end
      end
    elseif control.isButton() == 1 and button.home() == 1 and button.click() == 1 then
      die = true
    end
  until choice or die

  if not die then  -- Start selected game.
    cards = imgLib(choice.file, CARD_COLS, CARD_ROWS, choice.tile, choice.tile)
    backs = image.load(choice.back)
    total = board:new(choice.tile, BORDER_MIN, TILE_SPACE, cards, bkg, backs, BORDER, SHADOW)
    totalp = total / 2

    imgTaken = {}   -- Mark each with true when taken
    cardImg = {}    -- Holds tile numbers.
    imgNote = {}    -- Holds note numbers for each image.
    math.randomseed(os.time())  -- Randomize

    for i = 1, totalp, 1 do -- Select pairs.
      repeat
        randI = math.random (0, CARD_QTY - 1)
      until not imgTaken[randI]  -- Try again if used.
      imgTaken [randI] = true  -- Mark as used.
      cardImg [2 * i], cardImg[2 * i -1] = randI, randI  --Set two matching cards.
      imgNote [randI] = i + BASE_NOTE  -- Assign a sound to each image used.
    end

    for i = 1, SHUF_PASS, 1 do  --Shuffle the deck by repeated swapping.
      for j = 1, total, 1 do
        repeat
          randJ = math.random (1, total)
        until randJ ~= j
        cardImg[ j ], cardImg[ randJ ] = cardImg[ randJ ], cardImg[ j ]
      end
    end

    board:set()
    screen.update()

    moves = 0
    matches = 0
    moves = 0
    match = false
    while matches < totalp do
      control.read(1)
      if control.isTouch() == 1 and (touch.click() == 1 or touch.up() == 1) then
        xT, yT = touch.pos()
        tileT = board:pick (xT, yT)
        if tileT then  -- Was the click on a tile space?
          if cardImg [tileT] then  -- Is there a tile there?
            if match then  -- Is one card already up?
              if match ~= tileT then -- Make sure we didn't click the same one
                if soundON then
                  audio.beep(noteHz(imgNote[cardImg[tileT] ]), choice.wait)
                end
                board:up (tileT, cardImg[tileT])
                screen.update()
                os.wait(choice.wait)
                moves = moves + 1
                if cardImg[match] == cardImg[tileT] then  --Do they match
                  board:off(match)
                  board:off(tileT)
                  screen.update()
                  if soundON then
                    audio.beep(noteHz(imgNote[cardImg[tileT]]+12), choice.wait) -- Octave up means match!
                  end
                  cardImg[match] = nil
                  cardImg[tileT] = nil
                  match = false
                  matches = matches + 1
                else  --They don't match
                  board:down(match)
                  board:down(tileT)
                  screen.update()
                  if soundON then
                    audio.beep(noteHz(imgNote[cardImg[match] ]), choice.wait)
                  end
                  match = false
                end
              end
            else  --  No cards up yet.
              if soundON then
                audio.beep(noteHz(imgNote[cardImg[tileT] ]), choice.wait)
              end
              board:up (tileT, cardImg[tileT])
              screen.update()
              match = tileT  --Save to match
            end
          end
        end
      elseif control.isButton() == 1 and button.home() == 1 and button.click() == 1 then
        break
      end
    end

    if matches == totalp then
      if soundON then
        winSound = wav.load(WIN_WAV)
        winSound:play()
      end
      win = image.load("win.png")
      win:draw(0,0)
      text.size(24)
      text.color(BORDER)
      text.draw(0, 160, "  " .. moves .. " .", "center", 400)
      screen.update()
      if soundON then
        os.wait(winSound:time())
        winSound:close()
      end
      control.read(1)
      win:close()
    end
    cards:close()
  end
until die

menu:close()
bkg:close()